@@ -166,8 +166,8 @@ class ClustermgtdConfig:
         self.region = config.get("clustermgtd", "region")
         self.cluster_name = config.get("clustermgtd", "cluster_name")
         self.dynamodb_table = config.get("clustermgtd", "dynamodb_table")
-        self.master_private_ip = config.get("clustermgtd", "master_private_ip")
-        self.master_hostname = config.get("clustermgtd", "master_hostname")
+        self.head_node_private_ip = config.get("clustermgtd", "master_private_ip")
+        self.head_node_hostname = config.get("clustermgtd", "master_hostname")
         instance_name_type_mapping_file = config.get(
             "clustermgtd", "instance_type_mapping", fallback=self.DEFAULTS.get("instance_type_mapping")
         )
@@ -313,8 +313,8 @@ class ClusterManager:
             hosted_zone=config.hosted_zone,
             dns_domain=config.dns_domain,
             use_private_hostname=config.use_private_hostname,
-            master_private_ip=config.master_private_ip,
-            master_hostname=config.master_hostname,
+            head_node_private_ip=config.head_node_private_ip,
+            head_node_hostname=config.head_node_hostname,
             instance_name_type_mapping=config.instance_name_type_mapping,
         )
 
@@ -432,20 +432,21 @@ class ClusterManager:
 
     def _write_timestamp_to_file(self):
         """Write timestamp into shared file so compute nodes can determine if head node is online."""
-        with open(self._config.heartbeat_file_path, "w") as timestamp_file:
+        # Make clustermgtd heartbeat readable to all users
+        with open(os.open(self._config.heartbeat_file_path, os.O_WRONLY | os.O_CREAT, 0o644), "w") as timestamp_file:
             # Note: heartbeat must be written with datetime.strftime to convert localized datetime into str
             # datetime.strptime will not work with str(datetime)
             timestamp_file.write(datetime.now(tz=timezone.utc).strftime(TIMESTAMP_FORMAT))
 
     @staticmethod
     @retry(stop_max_attempt_number=2, wait_fixed=1000)
-    def _get_node_info_with_retry(nodes):
-        return get_nodes_info(nodes, command_timeout=10)
+    def _get_node_info_with_retry(nodes=""):
+        return get_nodes_info(nodes)
 
     @staticmethod
     @retry(stop_max_attempt_number=2, wait_fixed=1000)
     def _get_partition_info_with_retry():
-        return get_partition_info(command_timeout=10, get_all_nodes=True)
+        return get_partition_info(get_all_nodes=True)
 
     @staticmethod
     def _get_node_info_from_partition():
@@ -457,18 +458,22 @@ class ClusterManager:
         try:
             inactive_nodes = []
             active_nodes = []
-            partitions = ClusterManager._get_partition_info_with_retry()
+            ignored_nodes = []
+            partitions = {partition.name: partition for partition in ClusterManager._get_partition_info_with_retry()}
             log.debug("Partitions: %s", partitions)
-            for part in partitions:
-                # Get node info if partition has relevant node
-                # See SlurmNode.SLURM_RELEVANT_SINFO_STATES
-                if part.nodes:
-                    nodes = ClusterManager._get_node_info_with_retry(part.nodes)
-                    if "INACTIVE" in part.state:
-                        inactive_nodes.extend(nodes)
-                    else:
-                        active_nodes.extend(nodes)
+            nodes = ClusterManager._get_node_info_with_retry()
+            log.debug("Nodes: %s", nodes)
+            for node in nodes:
+                if not node.partitions or any(p not in partitions for p in node.partitions):
+                    # ignore nodes not belonging to any partition
+                    ignored_nodes.append(node)
+                elif any("INACTIVE" not in partitions[p].state for p in node.partitions):
+                    active_nodes.append(node)
+                else:
+                    inactive_nodes.append(node)
 
+            if ignored_nodes:
+                log.warning("Ignoring following nodes because they do not belong to any partition: %s", ignored_nodes)
             return active_nodes, inactive_nodes
         except Exception as e:
             log.error("Failed when getting partition/node states from scheduler with exception %s", e)
@@ -540,7 +545,7 @@ class ClusterManager:
         Instances returned will not contain instances previously terminated in _clean_up_inactive_partition
         """
         try:
-            return self._instance_manager.get_cluster_instances(include_master=False, alive_states_only=True)
+            return self._instance_manager.get_cluster_instances(include_head_node=False, alive_states_only=True)
         except Exception as e:
             log.error("Failed when getting instance info from EC2 with exception %s", e)
             raise ClusterManager.EC2InstancesInfoUnavailable
@@ -650,6 +655,9 @@ class ClusterManager:
                     nodes_failing_health_check.append(unhealthy_node.name)
         if nodes_failing_health_check:
             # Place unhealthy node into drain, this operation is idempotent
+            log.warning(
+                "Setting nodes failing health check type %s to DRAIN: %s", health_check_type, nodes_failing_health_check
+            )
             set_nodes_drain(nodes_failing_health_check, reason=f"Node failing {health_check_type}")
 
     def _update_static_nodes_in_replacement(self, slurm_nodes):
