@@ -15,9 +15,10 @@ import functools
 import json
 import logging
 import subprocess
-from datetime import timezone
+from datetime import datetime, timezone
 
 import boto3
+from botocore.config import Config
 from botocore.exceptions import ClientError
 
 from common.schedulers.slurm_commands import InvalidNodenameError, parse_nodename, update_nodes
@@ -105,8 +106,8 @@ class InstanceManager:
         hosted_zone=None,
         dns_domain=None,
         use_private_hostname=False,
-        master_private_ip=None,
-        master_hostname=None,
+        head_node_private_ip=None,
+        head_node_hostname=None,
         instance_name_type_mapping=None,
     ):
         """Initialize InstanceLauncher with required attributes."""
@@ -119,15 +120,17 @@ class InstanceManager:
         self._hosted_zone = hosted_zone
         self._dns_domain = dns_domain
         self._use_private_hostname = use_private_hostname
-        self._master_private_ip = master_private_ip
-        self._master_hostname = master_hostname
+        self._head_node_private_ip = head_node_private_ip
+        self._head_node_hostname = head_node_hostname
         self._instance_name_type_mapping = instance_name_type_mapping or {}
 
     def _clear_failed_nodes(self):
         """Clear and reset failed nodes list."""
         self.failed_nodes = []
 
-    def add_instances_for_nodes(self, node_list, launch_batch_size, update_node_address=True):
+    def add_instances_for_nodes(
+        self, node_list, launch_batch_size, update_node_address=True, all_or_nothing_batch=False
+    ):
         """Launch requested EC2 instances for nodes."""
         # Reset failed_nodes
         self._clear_failed_nodes()
@@ -137,7 +140,9 @@ class InstanceManager:
                 logger.info("Launching instances for slurm nodes %s", print_with_count(slurm_node_list))
                 for batch_nodes in grouper(slurm_node_list, launch_batch_size):
                     try:
-                        launched_instances = self._launch_ec2_instances(queue, instance_type, len(batch_nodes))
+                        launched_instances = self._launch_ec2_instances(
+                            queue, instance_type, len(batch_nodes), all_or_nothing_batch=all_or_nothing_batch
+                        )
                         if update_node_address:
                             assigned_nodes = self._update_slurm_node_addrs(list(batch_nodes), launched_instances)
                             try:
@@ -208,8 +213,8 @@ class InstanceManager:
                         Item={
                             "Id": nodename,
                             "InstanceId": instance.id,
-                            "MasterPrivateIp": self._master_private_ip,
-                            "MasterHostname": self._master_hostname,
+                            "MasterPrivateIp": self._head_node_private_ip,
+                            "MasterHostname": self._head_node_hostname,
                         }
                     )
 
@@ -240,7 +245,13 @@ class InstanceManager:
             # Submit calls to change_resource_record_sets in batches of 500 elements each.
             # change_resource_record_sets API call has limit of 1000 changes,
             # but the UPSERT action counts for 2 calls
-            route53_client = boto3.client("route53", region_name=self._region, config=self._boto3_config)
+            # Also pick the number of retries to be the max between the globally configured one and 3.
+            # This is done to address Route53 API throttling without changing the configured retries for all API calls.
+            configured_retry = self._boto3_config.retries.get("max_attempts", 0) if self._boto3_config.retries else 0
+            boto3_config = self._boto3_config.merge(
+                Config(retries={"max_attempts": max([configured_retry, 3]), "mode": "standard"})
+            )
+            route53_client = boto3.client("route53", region_name=self._region, config=boto3_config)
             changes_batch_size = 500
             for changes_batch in grouper(changes, changes_batch_size):
                 route53_client.change_resource_record_sets(
@@ -268,33 +279,37 @@ class InstanceManager:
 
         return instances_to_launch
 
-    def _launch_ec2_instances(self, queue, instance_type, current_batch_size, best_effort=True):
+    def _launch_ec2_instances(self, queue, instance_type, current_batch_size, all_or_nothing_batch=False):
         """Launch a batch of ec2 instances."""
-        ec2_client = boto3.client("ec2", region_name=self._region, config=self._boto3_config)
-        result = ec2_client.run_instances(
-            # If best_effort scaling, set MinCount=1
-            # so run_instances call will succeed even if entire count cannot be satisfied
-            # Otherwise set MinCount=current_batch_size so run_instances will fail unless all are launched
-            MinCount=1 if best_effort else current_batch_size,
-            MaxCount=current_batch_size,
-            # LaunchTemplate is different for every instance type in every queue
-            # LaunchTemplate name format: {cluster_name}-{queue_name}-{instance_type}
-            # Sample LT name: hit-queue1-c5.xlarge
-            LaunchTemplate={
-                "LaunchTemplateName": f"{self._cluster_name}-{queue}-{instance_type}",
-                "Version": "$Latest",
-            },
-        )
-
-        return [
-            EC2Instance(
-                instance_info["InstanceId"],
-                instance_info["PrivateIpAddress"],
-                instance_info["PrivateDnsName"].split(".")[0],
-                instance_info["LaunchTime"],
+        try:
+            ec2_client = boto3.client("ec2", region_name=self._region, config=self._boto3_config)
+            result = ec2_client.run_instances(
+                # If not all_or_nothing_batch scaling, set MinCount=1
+                # so run_instances call will succeed even if entire count cannot be satisfied
+                # Otherwise set MinCount=current_batch_size so run_instances will fail unless all are launched
+                MinCount=1 if not all_or_nothing_batch else current_batch_size,
+                MaxCount=current_batch_size,
+                # LaunchTemplate is different for every instance type in every queue
+                # LaunchTemplate name format: {cluster_name}-{queue_name}-{instance_type}
+                # Sample LT name: hit-queue1-c5.xlarge
+                LaunchTemplate={
+                    "LaunchTemplateName": f"{self._cluster_name}-{queue}-{instance_type}",
+                    "Version": "$Latest",
+                },
             )
-            for instance_info in result["Instances"]
-        ]
+
+            return [
+                EC2Instance(
+                    instance_info["InstanceId"],
+                    instance_info["PrivateIpAddress"],
+                    instance_info["PrivateDnsName"].split(".")[0],
+                    instance_info["LaunchTime"],
+                )
+                for instance_info in result["Instances"]
+            ]
+        except ClientError as e:
+            logger.error("Failed RunInstances request: %s", e.response.get("ResponseMetadata").get("RequestId"))
+            raise
 
     def delete_instances(self, instance_ids_to_terminate, terminate_batch_size):
         """Terminate corresponding EC2 instances."""
@@ -307,6 +322,9 @@ class InstanceManager:
                     InstanceIds=list(instances),
                 )
             except ClientError as e:
+                logger.error(
+                    "Failed TerminateInstances request: %s", e.response.get("ResponseMetadata").get("RequestId")
+                )
                 logger.error("Failed when terminating instances %s with error %s", print_with_count(instances), e)
 
     @log_exception(logger, "getting health status for unhealthy EC2 instances", raise_on_error=True)
@@ -351,7 +369,7 @@ class InstanceManager:
         return list(instance_health_states.values())
 
     @log_exception(logger, "getting cluster instances from EC2", raise_on_error=True)
-    def get_cluster_instances(self, include_master=False, alive_states_only=True):
+    def get_cluster_instances(self, include_head_node=False, alive_states_only=True):
         """Get instances that are associated with the cluster."""
         ec2_client = boto3.client("ec2", region_name=self._region, config=self._boto3_config)
         paginator = ec2_client.get_paginator("describe_instances")
@@ -360,7 +378,7 @@ class InstanceManager:
         }
         if alive_states_only:
             args["Filters"].append({"Name": "instance-state-name", "Values": list(EC2_INSTANCE_ALIVE_STATES)})
-        if not include_master:
+        if not include_head_node:
             args["Filters"].append({"Name": "tag:aws-parallelcluster-node-type", "Values": ["Compute"]})
         response_iterator = paginator.paginate(PaginationConfig={"PageSize": BOTO3_PAGINATION_PAGE_SIZE}, **args)
         filtered_iterator = response_iterator.search("Reservations[].Instances[]")
@@ -388,13 +406,13 @@ def time_is_up(initial_time, current_time, grace_time):
     """Check if timeout is exceeded."""
     # Localize datetime objects to UTC if not previously localized
     # All timestamps used in this function should be already localized
-    # Assume timestamp was taken from UTC is there is no localization info
+    # Assume timestamp was taken from system local time if there is no localization info
     if not initial_time.tzinfo:
         logger.warning("Timestamp %s is not localized. Please double check that this is expected, localizing to UTC.")
-        initial_time = initial_time.replace(tzinfo=timezone.utc)
+        initial_time = initial_time.astimezone(tz=timezone.utc)
     if not current_time.tzinfo:
         logger.warning("Timestamp %s is not localized. Please double check that this is expected, localizing to UTC")
-        current_time = current_time.replace(tzinfo=timezone.utc)
+        current_time = current_time.astimezone(tz=timezone.utc)
     time_diff = (current_time - initial_time).total_seconds()
     return time_diff >= grace_time
 
@@ -411,3 +429,35 @@ def retrieve_instance_type_mapping(file_path):
             e,
         )
         raise
+
+
+def _get_clustermgtd_heartbeat(clustermgtd_heartbeat_file_path):
+    """Get clustermgtd's last heartbeat."""
+    with open(clustermgtd_heartbeat_file_path, "r") as timestamp_file:
+        # Note: heartbeat must be written with datetime.strftime to convert localized datetime into str
+        # datetime.strptime will not work with str(datetime)
+        # Example timestamp written to heartbeat file: 2020-07-30 19:34:02.613338+00:00
+        return datetime.strptime(timestamp_file.read().strip(), TIMESTAMP_FORMAT)
+
+
+def _expired_clustermgtd_heartbeat(last_heartbeat, current_time, clustermgtd_timeout):
+    """Test if clustermgtd heartbeat is expired."""
+    if time_is_up(last_heartbeat, current_time, clustermgtd_timeout):
+        logger.error(
+            "Clustermgtd has been offline since %s. Current time is %s. Timeout of %s seconds has expired!",
+            last_heartbeat,
+            current_time,
+            clustermgtd_timeout,
+        )
+        return True
+    return False
+
+
+def is_clustermgtd_heartbeat_valid(current_time, clustermgtd_timeout, clustermgtd_heartbeat_file_path):
+    try:
+        last_heartbeat = _get_clustermgtd_heartbeat(clustermgtd_heartbeat_file_path)
+        logger.info("Latest heartbeat from clustermgtd: %s", last_heartbeat)
+        return not _expired_clustermgtd_heartbeat(last_heartbeat, current_time, clustermgtd_timeout)
+    except Exception as e:
+        logger.error("Unable to retrieve clustermgtd heartbeat with exception: %s", e)
+        return False
